/***
 * This is the main file for the sysdiag utility.
 *
 *
 * Copyright (C) 2003 International Business Machines, Inc.
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation version 2 of the License.
 *
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License along
 *	with this program; if not, write to the Free Software Foundation, Inc.,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * 	need api in libsysfs to get one attribute of one particular device.
 *	also an api to get all attributes of one device.
 *
 */

#include "sysdiag.h"


/**
 * displays long usage.
 */
int usage()
{

	fprintf(stdout, "sysdiag [<options> [dev-name]]\n\n");
	fprintf(stdout, " options:\n");
	fprintf(stdout, "\t-h\t\t\tusage\n");
	fprintf(stdout, "\t-d\t\t\tShow diagnostics capabilites\n");
	fprintf(stdout, "\t-R\t\t\tReload the database.\n\n");
	fprintf(stdout, "\t-i\t\t\trun device inquiry cmd\n");
	fprintf(stdout, "\t-s\t\t\tshow device statistics\n");
	fprintf(stdout, "\t-t\t\t\trun diagnostic test cmd\n");
	fprintf(stdout, "\t-l\t\t\tshow logs\n");
	fprintf(stdout, "\t-L\t\t\tshow logs from saved error logs\n");
	fprintf(stdout, "\t-x\t\t\trun diag on all devices(use with one of \n\t\t\t\tthe diagnostic commands)\n\n");
	fprintf(stdout, "\t-p\t\t\tstop device before running diag\n");
	fprintf(stdout, "\t-b <bus-name>\t\tspecify bus to use\n");
	fprintf(stdout, "\t-c <class-name>\t\tspecify class to use\n");
	fprintf(stdout, "\t-r <root-dev>\t\tspecify root device to show\n");
//	fprintf(stdout, "\t-B <block-device-name>\tspecify block device to use\n");
	fprintf(stdout, "\t-v\t\t\tmore verbose\n");
	fprintf(stdout, "\ndev-name:\t\teth0,XXXX:XX:XX.X (busid), X:X:X:X (scsi id) etc.\n");
	fprintf(stdout, "bus-name:\t\tpci, scsi, etc.\n");
	fprintf(stdout, "class-name:\t\tnet, scsi_host, etc.\n");
	fprintf(stdout, "root-dev:\t\tpci0000:00, legacy, etc.\n");
	return 0;
}



void show_diag_info(char *device)
{
	struct diags *curr = NULL;
	struct dlist *list = NULL;

	list = (struct dlist *)sysdiag_get_diag_list();
	if ( list == NULL) {
		fprintf(stderr, "list is null\n");
		return ;
	}

	dlist_for_each_data(list, curr, struct diags) {
		if (device != NULL) {
			if (strncasecmp(curr->name, device, 
						strlen(device)) != 0) {
				continue;
			}
		}
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Diagnostics for class %s:\n", curr->name);
		fprintf(stdout, "-------------------------------------\n");
		if (curr->test != NULL) {
			fprintf(stdout, "Test diag option      :  -t"
					"\t%s\n", curr->test);
		}
		if (curr->inq != NULL) {
			fprintf(stdout, "Inquiry Diag option   :  -i"
					"\t%s\n", curr->inq);
		}
		if (curr->stat != NULL) {
			fprintf(stdout, "Statistic Diag option :  -s"
					"\t%s\n", curr->stat);
		}
	}
		
}


/**
 * This will setup the various sysfs paths.
 */
int setup_sysfs_paths()
{
	memset(SYSFS_MOUNT_PATH, 0, SYSFS_PATH_MAX);
	if ( sysfs_get_mnt_path(SYSFS_MOUNT_PATH, SYSFS_PATH_MAX) != 0) {
		fprintf(stderr,"sysfs not supported on your system\n");
		return -1;
	}
	return 0;

}

int sysfs_get_def_scsi_dev()
{
	struct sysfs_bus *bus = NULL;
	struct sysfs_device *dev = NULL, *cur = NULL;
	
	if ((bus = sysfs_open_bus("scsi")) == NULL) {
		fprintf(stdout, "failed opeing bus :scsi\n");
		return -1;
	}
	if (bus->devices == NULL) {
		fprintf(stdout, "no scsi device present\n");
		sysfs_close_bus(bus);
		return -1;
	}
	dlist_for_each_data(bus->devices, cur, struct sysfs_device)
		dev = cur;
	
	strncpy(DEVICE_NAME, dev->bus_id, strlen(dev->bus_id));
	sysfs_close_bus(bus);
	return -1;	

}



char * get_scsi_quad(char *device)
{
	char block_key[SYSFS_NAME_LEN], key[SYSFS_NAME_LEN];
	char block_dev_key[SYSFS_NAME_LEN];
	char *dev_num = NULL, *data = NULL, *str = NULL;
	char *datastr = NULL;
	int num = 0, j;

	if (DEVICE_NAME[0] == 0) {
		sysfs_get_def_scsi_dev();
		if (DEVICE_NAME[0] == 0) {
			return NULL;
		}
	}
	
	memset(block_key, 0, SYSFS_NAME_LEN);
	strncpy(block_key, SYSDIAG_CLASS_STR, strlen(SYSDIAG_CLASS_STR));
	strncat(block_key, "#devices#", 9);
	strncat(block_key, SYSDIAG_BLOCK_STR, strlen(SYSDIAG_BLOCK_STR));
	
	data = sysdiag_get_record(block_key);
	num = sysdiag_db_num_of_elements(data);
	
	memset(block_dev_key, 0, SYSFS_NAME_LEN);
	strncpy(block_dev_key, SYSDIAG_BLOCK_STR, strlen(SYSDIAG_BLOCK_STR));
	strncat(block_dev_key, "#", 1);
	
	for (j = 1; j <= num; j++) {
		str = sysdiag_db_get_str(data, j);
		if (str == NULL) {
			free(data);
			return NULL;
		}
		memset(key, 0, SYSFS_NAME_LEN);
		strncpy(key, block_dev_key, strlen(block_dev_key));
		strncat(key, str, strlen(str));
		datastr = sysdiag_get_record(key);
		if (datastr == NULL)
			continue;
		if ((strcmp(datastr, device) != 0) && 
				(strcmp(str, device) != 0)) {
			free(datastr);
			continue;
		}
		free(data);
		return (char *)str;
	}
	free(data);
	return NULL;
}


char *get_def_scsi_host_dev(char *device)
{
	struct sysfs_class_device *classdev = NULL;
	struct sysfs_directory *mydir = NULL;
	char *dev_num = NULL;

	if ((classdev = sysfs_open_class_device_by_name("scsi_host", 
					device)) == NULL) {
		fprintf(stdout, "error opening class device %s\n", device);
		return NULL;
	}
	if (classdev->sysdevice == NULL) {
		fprintf(stdout, "no device for class\n");
		return NULL;
	}
	
	if (classdev->sysdevice->directory->subdirs == NULL) {
		fprintf(stdout, "no device on host : %s\n", device);
		return NULL;
	}
	dlist_for_each_data(classdev->sysdevice->directory->subdirs, mydir, 
			struct sysfs_directory) {
		dev_num = calloc(1, strlen(mydir->name));
		strncpy(dev_num, mydir->name, strlen(mydir->name));
		break;
	}
	
	sysfs_close_class_device(classdev);
	dev_num = get_scsi_quad(dev_num);
	return dev_num;
}

/**
 *
 */
char *make_diag_path(char *path, char *dev_num)
{
	char *ptr = NULL;
	char *diag_path = NULL;
	int size = 0;

	size = strlen(dev_num);
	if ((ptr = strstr(path, "<dev>")) != NULL) {
		size += strlen(path);
		size-=5;
		diag_path = (char *)calloc(size+1, sizeof(char));
		strncpy(diag_path, path, (ptr - path));
		strncat(diag_path, dev_num, strlen(dev_num));
		ptr+=5;
		strncat(diag_path, ptr, strlen(ptr));
	}
	else {
		size += strlen(path);
		diag_path = (char *)calloc(size+1, sizeof(char));
		strncpy(diag_path, path, strlen(path));
		strncat(diag_path, " ", 1);
		strncat(diag_path, dev_num, strlen(dev_num));
	}
	return (char *)diag_path;

}


/**
 *  
 */
int show_err_logs(char *dev_num)
{
	char *data = NULL;
	char *path = NULL;
	int flag = 0;


	if (geteuid() != 0) {
		fprintf(stderr, "need to log in as root!\n");
		return -1;
	}

#ifdef HAVE_EVLVIEW 
	data = sysdiag_get_record(SYSDIAG_EVLOG_KEY);
	if ((path = (char *)make_diag_path(data, dev_num)) != NULL) {
		fprintf(stdout, "\n---------------------------------------------\n");
		fprintf(stdout, "Error logs from evlog for Device : %s\n", dev_num);
		fprintf(stdout, "----------------------------------------------\n");
		fflush(stdout);
		system(path);
	}
	free(data);
	free(path);
#endif
	path = NULL;
	if (DIAG_OPTIONS & DIAG_SHOW_OLD_LOGS)
		flag = 1;
	path = (char *)show_logs(dev_num, flag);
	fprintf(stdout, "\n-------------------------------------\n");
	fprintf(stdout, "Error logs for Device : %s\n", dev_num);
	fprintf(stdout, "-------------------------------------\n");
	fflush(stdout);
	system(path);
	
	if (path != NULL)
		free(path);

}

/**
 * This function based on the option will call the relavent
 * sg3utils applications.
 * dev_num is in sgX format.
 */
int run_scsi_diag(char *device)
{
	char *path = NULL;
	char num[10];
	int size = 0;
	char dev_num[SYSFS_NAME_LEN];
	struct diags *curr_diag = NULL;

	struct dlist *list = NULL;

	memset(dev_num, 0, SYSFS_NAME_LEN);
	strncpy(dev_num, "/dev/",5);
	strncat(dev_num, device, strlen(device));

	list = (struct dlist *)sysdiag_get_diag_list();
	if (list == NULL) {
		fprintf(stdout, "no diagnostic information is available\n");
		return -1;
	}
	dlist_for_each_data(list, curr_diag, struct diags) {
		if (curr_diag != NULL) {
			if (strcmp(curr_diag->name, "scsi_device") != 0)
				continue;
			else 
				break;
		}
		return -1;
	}
	
	size = strlen(dev_num);
	if (DIAG_OPTIONS & DIAG_INQ) {
		if (curr_diag->inq != NULL)
			path = make_diag_path(curr_diag->inq, dev_num);
		else {
			fprintf(stderr, 
				"no inquiry diagnostics available for %s\n",
				dev_num);
			return -1;
		}
		
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Device inquiry on : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		fflush(stdout);
		system(path);
	}
	if (DIAG_OPTIONS & DIAG_STAT) {
		if (curr_diag->stat != NULL)
			path = make_diag_path(curr_diag->stat, dev_num);
		else {
			fprintf(stderr, 
				"no statistics diagnostics available for %s\n",
				dev_num);
			return -1;
		}

		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Device statistics of : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		fflush(stdout);
		system(path);
	}
	if (DIAG_OPTIONS & DIAG_TEST) {
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Testing Device : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		if (curr_diag->test != NULL)
			path = make_diag_path(curr_diag->test, dev_num);
		else {
			fprintf(stderr, 
				"no test diagnostics available for %s\n",
				dev_num);
			return -1;
		}

		if (DIAG_OPTIONS & DIAG_PAUSE) {
			char *down = NULL;
			size = strlen(dev_num) + 15;
			down = (char *)calloc(size, sizeof(char));
			strncpy(down, "sg_start  ", 10);
			strncat(down, dev_num, strlen(dev_num));
			strncat(down, "  0", 3);
			fprintf(stdout, "Bringing down device:%s\n", dev_num);
			fflush(stdout);
			system(down);
			free(down);
		}
		fflush(stdout);
		system(path);
		if (DIAG_OPTIONS & DIAG_PAUSE) {
			char *up = NULL;
			size = strlen(dev_num) + 15;
			up = (char *)calloc(size, sizeof(char));
			strncpy(up, "sg_start  ", 10);
			strncat(up, dev_num, strlen(dev_num));
			strncat(up, "  1", 3);
			fprintf(stdout, "Bringing up device:%s\n", dev_num);
			fflush(stdout);
			system(up);
			free(up);
		}
	}
	if (DIAG_OPTIONS & DIAG_LOGS) {
		show_err_logs(device);
	}
	if (path != NULL)
		free(path);
	return 0;
}


/**
 * run_scsi_diag_all().
 * this runs diag on all the scsi devices.
 * if DEVICE_NAME is not null
 *	then we have to run diag on all devices on the that
 * 	host bus adapter, i.e. host0, host1 etc. For this we
 *	go to /sys/class/scsi_host/host0/device directory and
 *	and run diag on all the devices there.
 *
 * else
 *	we go to the sys/bus/scsi/devices/ dir and run diag on
 *	all devices there.
 *
 *
 *
 */
int run_scsi_diag_all()
{
	struct sysfs_class_device *clsdev = NULL;
	struct sysfs_class *cls = NULL;
	char *dev_num = NULL, name[SYSFS_NAME_LEN];

	struct sysfs_class_device *classdev = NULL;
	struct sysfs_device *dev = NULL;
	struct sysfs_directory *mydir = NULL;

	if (DEVICE_NAME[0] == 0) {
		if ((cls = sysfs_open_class("scsi_device")) == NULL) {
			fprintf(stderr, "error opening class: scsi_device");
			return -1;
		}
	}
	else {
		if ((classdev = sysfs_open_class_device_by_name("scsi_host", 
						DEVICE_NAME)) == NULL) {
			fprintf(stdout, "error opening class device %s\n", 
					DEVICE_NAME);
			return -1;
		}
		if (classdev->sysdevice == NULL) {
			fprintf(stdout, "no device for this class\n");
			return -1;
		}
		if (classdev->sysdevice->directory->subdirs == NULL) {
			fprintf(stdout, "no device on host: %s\n", DEVICE_NAME);
			return -1;
		}
		dlist_for_each_data(classdev->sysdevice->directory->subdirs, 
				mydir, 	struct sysfs_directory) {
			dev_num = calloc(1, strlen(mydir->name));
			strncpy(dev_num, mydir->name, strlen(mydir->name));
			dev_num = get_scsi_quad(dev_num);
			if (dev_num == NULL) {
				sysfs_close_class(cls);
				return -1;
			}
			run_scsi_diag(dev_num);
		}
		sysfs_close_class_device(classdev);
		return 0;
	}
	if (cls->devices != NULL) {
		dlist_for_each_data(cls->devices, clsdev, 
				struct sysfs_class_device) {
			memset(name, 0, SYSFS_NAME_LEN);
			strncpy(name, clsdev->name, strlen(clsdev->name));
			dev_num = get_scsi_quad(name);
			if (dev_num == NULL) {
				sysfs_close_class(cls);
				return -1;
			}
			run_scsi_diag(dev_num);
		}
	}
	free(dev_num);
	sysfs_close_class(cls);
	return 0;
}


/**
 * this will call the ethtool with the relevant options.
 * the dev_num is in ethx format.
 */
int run_net_diag(char *dev_num)
{
	char *path = NULL;
	int size = 0;
	struct diags *curr_diag = NULL;
	struct dlist *list = NULL;

	list = (struct dlist *)sysdiag_get_diag_list();
	if (list == NULL) {
		fprintf(stdout, "no diagnostic information is available\n");
			return -1;
	}
	
	dlist_for_each_data(list, curr_diag, struct diags) {
		if (curr_diag!= NULL) {
			if (strcmp(curr_diag->name, "net") == 0)
				break;
		}
		return -1;
	}
	size  = strlen(dev_num);

	if (DIAG_OPTIONS & DIAG_INQ) {
		if (curr_diag->inq != NULL) 
			path = make_diag_path(curr_diag->inq, dev_num);
		else {
			fprintf(stderr, 
				"no inquiry diagnostics available for %s\n",
				dev_num);
			return -1;
		}
		
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Device inquiry on : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		fflush(stdout);
		system(path);
	}
	if (DIAG_OPTIONS & DIAG_STAT) {
		if (curr_diag->stat != NULL)
			path = make_diag_path(curr_diag->stat, dev_num);
		else {
			fprintf(stderr, 
				"no statistics diagnostics available for %s\n",
				dev_num);
			return -1;
		}
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Device statistics of : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		fflush(stdout);
		system(path);
	}
	if (DIAG_OPTIONS & DIAG_TEST) {
		if (curr_diag->test != NULL)
			path = make_diag_path(curr_diag->test, dev_num);
		else {
			fprintf(stderr, 
				"no test diagnostics available for %s\n",
				dev_num);
			return -1;
		}
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Testing Device : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		if (DIAG_OPTIONS & DIAG_PAUSE) {
			char *down = NULL;
			size = strlen(dev_num) + 15;
			down = (char *)calloc(size, sizeof(char));
			strncpy(down, "ifconfig ", 9);
			strncat(down, dev_num, strlen(dev_num));
			strncat(down, " down", 5);
			fprintf(stdout, "Bringing down device:%s\n", dev_num);
			fflush(stdout);
			system(down);
			free(down);
		}
		fflush(stdout);
		system(path);

		if (DIAG_OPTIONS & DIAG_PAUSE) {
			char *up = NULL;
			size = strlen(dev_num) + 13;
			up = (char *)calloc(size, sizeof(char));
			strncpy(up, "ifconfig ", 9);
			strncat(up, dev_num, strlen(dev_num));
			strncat(up, " up",3);
			fprintf(stdout, "Bringing up device:%s\n", dev_num);
			fflush(stdout);
			system(up);
			free(up);
		}
	}
	if (DIAG_OPTIONS & DIAG_LOGS) {
		show_err_logs(dev_num);
	}
	if (path != NULL)
		free(path);
	return 0;
}


/**
 * run_net_diag_all():
 * This will work with only the class option,
 * goto the sys/class/net dir and run diag on all devices
 * there, check if the name of the device is in the ethX format,
 * to take care of the lo and dummy driver case.
 *
 *
 */
int run_net_diag_all(void)
{
	struct sysfs_class *class = NULL;
	struct sysfs_class_device *child = NULL;
	char name[SYSFS_NAME_LEN];
	
	if ((class = sysfs_open_class("net")) == NULL) {
		fprintf(stderr, "error opening class: net\n");
		return 1;
	}
	
	if (class->devices != NULL) {
		dlist_for_each_data(class->devices, child, 
				struct sysfs_class_device) {
			if (strncmp(child->name, "eth", 3) != 0) {
				continue;
			}
			memset(name, 0, SYSFS_NAME_LEN);
			strncpy(name, child->name, strlen(child->name));
			run_net_diag(child->name);
		}
	}
	sysfs_close_class(class);
	return 0;
}


/**
 * given device-name it will check if it is a valid device name and
 * also get the valid device name in case the format is diff than
 * ethX.
 * in case of net class the device name will be in ethX format.
 * we just need to validate it.
 *
 * the network diags are only with the class option.
 * option -c net -D ethX or 0000:00:00.0, or 00:00.0
 * but doesnot work with the pci option, well we can make it with
 * the pci option, since right now only the device we will run diag
 * thro the bus option will be a network device not scsi device.
 *
 * for pci only network devices will be test. If a scsi device
 * bus_id is passed in -b pci option it will not work.
 * Need to pass that with -b scsi option.
 *
 * in case of pci, the format will be 0000:00:00.0 and also it can
 * so we need to check if that device is a network device.
 * check of class of device is 02.
 * in sysfs format for class is 020000, firt byte is the class.
 * so take first two chars and cmp with 02.
 *
 */
char *get_net_dev_num(char * dev_name)
{
	char *path = NULL, *name = NULL;
	struct sysfs_class_device *child = NULL;
	struct sysfs_class *cls = NULL;

	if ((cls = sysfs_open_class("net")) == NULL) {
		fprintf(stderr, "error opening class: net\n");
		return NULL;
	}
	
	if (cls->devices != NULL) {
		name = malloc(SYSFS_NAME_LEN);
		dlist_for_each_data(cls->devices, child, 
				struct sysfs_class_device) {
			if (strncmp(child->name, "eth", 3) != 0) {
				continue;
			}
			memset(name, 0, SYSFS_NAME_LEN);
			strncpy(name, child->name, strlen(child->name));
			if (strncmp(name, "eth",3) != 0)
				continue;
			if (dev_name == NULL && DEVICE_NAME[0] == 0) {
				sysfs_close_class(cls);
				return name;
			}
			if (strncmp(name, DEVICE_NAME, 
						strlen(DEVICE_NAME)) == 0) {
				sysfs_close_class(cls);
				return name;
			}
		}
	}
	sysfs_close_class(cls);
	return NULL;
}


char *get_net_bus_dev(char *dev_name)
{

	struct sysfs_class_device *clsdev = NULL, *child = NULL;
	struct sysfs_class *cls = NULL;
	struct sysfs_device *dev = NULL;

	if ((cls = sysfs_open_class("net")) == NULL) {
		fprintf(stderr, "error opening class: net\n");
		return NULL;
	}
	
	dev_name = malloc(SYSFS_NAME_LEN);
	if (cls->devices != NULL) {
		dlist_for_each_data(cls->devices, child, 
				struct sysfs_class_device) {
			if (strncmp(child->name, "eth", 3) != 0) {
				continue;
			}
			
			memset(dev_name, 0, SYSFS_NAME_LEN);
			strncpy(dev_name, child->name, strlen(child->name));
			
			if (child->sysdevice != NULL) {
				dev = child->sysdevice;
				if (strncmp(dev->bus_id, DEVICE_NAME, 
						strlen(DEVICE_NAME)) == 0) {
					sysfs_close_class(cls);
					return dev_name;
				}
			}
		}
	}

	free(dev_name);
	sysfs_close_class(cls);	
	return NULL;
}



char *get_net_dev(char *dev_num)
{

	if (DEVICE_NAME != NULL) {
		if (DEVICE_NAME[2] == ':' || DEVICE_NAME[4] == ':') {
			if ((dev_num = get_net_bus_dev(dev_num)) == NULL) {
				return NULL;
			}
			return (char *)dev_num;
		}
	}
	if ((dev_num = get_net_dev_num(dev_num)) == NULL) {
		fprintf(stderr, "failed to get network device name\n");
		return NULL;
	}
	return (char *)dev_num;

}


int run_diag_bus()
{
	char *dev_num = NULL;
	char *path = NULL;
	
	if (strncmp(BUS_NAME, "scsi", 4) == 0) {
		if (DIAG_OPTIONS & DIAG_ALL) {
			if (DIAG_OPTIONS & DIAG_LOGS) {
				show_err_logs("scsi");
			}
			return (run_scsi_diag_all());
		}
		if ((dev_num = get_scsi_quad((char *)DEVICE_NAME)) == NULL) {
			fprintf(stderr, "failed to get scsi device name\n");
			return 1;
		}
		run_scsi_diag(dev_num);
		free(dev_num);
		return 0;
	}
	if (strncmp(BUS_NAME, "pci", 3) == 0) {
		if (DIAG_OPTIONS & DIAG_ALL)
			return (run_net_diag_all());
		if ((dev_num = get_net_dev(dev_num)) == NULL)
			return 1;
		run_net_diag(dev_num);
		free(dev_num);
		return 0;
	}
	return 1;
}

/**
 * currently the diagnostics are run on only the net, scsi_host, scsi_device
 * and block classes. 
 * 
 */
int run_diag_class()
{
	char *dev_num = NULL;
	char *path = NULL;

	if (strncmp(CLASS_NAME, "net", 3) == 0) {
		if (DIAG_OPTIONS & DIAG_ALL) {
			return (run_net_diag_all());
		}
		if ((dev_num = get_net_dev(dev_num)) == NULL) 
			return 1;
		run_net_diag(dev_num);
		free(dev_num);
		return 0;
	} else {
		if ((strstr(CLASS_NAME, "scsi") == NULL) &&
				(strcmp(CLASS_NAME, "block") != 0))
			return 1;

		if ((strncmp(CLASS_NAME, "scsi_device", 11) == 0) || 
				(strncmp(CLASS_NAME, "scsi_host", 9) == 0)) {
			if (DIAG_OPTIONS & DIAG_ALL) {
				if (DIAG_OPTIONS & DIAG_LOGS) {
					show_err_logs("scsi");
				}
				return (run_scsi_diag_all());
			}
		}
		if (strncmp(DEVICE_NAME, "host", 4) == 0) {
			if ((dev_num = get_def_scsi_host_dev(DEVICE_NAME))
					== NULL) 
				return 1;
		} else {
			if ((dev_num = get_scsi_quad((char *)DEVICE_NAME)) 
					== NULL) {
				fprintf(stderr, 
					"failed to get scsi device name\n");
				return 1;
			}
		}
		run_scsi_diag(dev_num);
		free(dev_num);
		return 0;
	}
	return 1;
}


/**
 * This function will parse the dev-class and dev-num and call
 * the appropriate function to show the device information.
 *
 * for sysdiag -i net
 * this function will look for the firt device in the net dir or the
 * /sysfs/class. and run ethtool -i on that.
 *
 * For ethernet device, in the /sys/class/net class, we look for the
 * ethX device and do the diag on it.
 *
 * For scsi device, in the /sys/class/scsi_host  dir, we go thro the
 * hostX dir and then in the device dir look at the 0:0:0:0 and then
 * look the link info for block and then see device type, map it to
 * the sgX device and then do the diagnostics on that.
 *
 * And if the scsi_host class is passed then it will call the sg_inq
 * function to show the information for that scsi_device.
 *
 *
 * Format for dev-info is
 * #sysdiag -i [dev-class] [dev-num]
 *
 * dev-class	: scsi_host, net.
 * dev-num	: eth0, sg0, 0:0:0:0.
 *
 * format for run_diag is::
 * sysdiag -i -b pci -D 0000:00:00.0
 * sysdiag -i -b pci -A
 * sysdiag -i -c net [-D eth0] or -A
 * sysdiag -i -c scsi_host [-D 0:0:0:0 or sg0 or sda] or -A
 *
 * sysdiag -a eth0
 * sysdiag -i eth0
 * sysdiag net eth0
 * 
 *
 * This function will parse the dev-class and dev-num and call
 * the appropriate function to run diagnostics on the appropriate
 * device.
 */
int run_diag()
{
	char *path = NULL;

	if (DIAG_OPTIONS & DIAG_CLASS) {
		return run_diag_class();
	}else if (DIAG_OPTIONS & DIAG_BUS) {
		return run_diag_bus();
	}
	if (DIAG_OPTIONS & DIAG_ALL) {
		if (DIAG_OPTIONS & DIAG_LOGS) {
			show_err_logs("scsi");
			DIAG_OPTIONS &= ~DIAG_LOGS;
			run_scsi_diag_all();
			DIAG_OPTIONS |= DIAG_LOGS;
		} else 
			run_scsi_diag_all();
		return run_net_diag_all();
	}
	return 1;
}

/**
 *  
 *
 */
int search_class_devices(struct dlist *clsdev, char *device)
{
	struct sysfs_class_device *cur = NULL;
	
	dlist_for_each_data(clsdev, cur, struct sysfs_class_device) {
		if (strcmp(cur->name, device) == 0) {
			strncpy(DEVICE_NAME, cur->name, strlen(cur->name));
			return 0;
		}
	}
	return -1;
}

/**
 *
 *
 */
int search_drivers(struct dlist *drvlist, char *device)
{
	struct sysfs_driver *cur = NULL;
	
	dlist_for_each_data(drvlist, cur, struct sysfs_driver) {
		if (strcmp(cur->name, device) == 0) {
			strncpy(DRIVER_NAME, cur->name, strlen(cur->name));
			return 0;
		}	
	}
}

/**
 *
 */

int search_devices(struct dlist *devlist, char *device)
{
	struct sysfs_device *cur = NULL;
	
	if (devlist != NULL) {
		dlist_for_each_data(devlist, cur, struct sysfs_device) {
			if (strcmp(cur->bus_id, device) == 0) {
				strncpy(DEVICE_NAME, cur->bus_id, 
						strlen(cur->bus_id));
				return 0;
			}
			if (cur->children != NULL) {
				if (search_devices(cur->children, device) == 0) {
					return 0;
				}
			}
		}
	}
	return -1;
}

/**
 *
 */
int new_search_root_incomplete(char *root_name, char *device)
{
	char *data = NULL;
	char key[SYSFS_NAME_LEN];
	int num = 0;
	
	memset(key, 0, SYSFS_NAME_LEN);
	strncpy(key, SYSFS_DEVICES_NAME, strlen(SYSDIAG_CLASS_STR));
	strncat(key, "#", 1);
	strncat(key, SYSFS_DEVICES_NAME, strlen(SYSFS_DEVICES_NAME));
	strncat(key, "#", 1);
	strncat(key, root_name, strlen(root_name));
	num = 0;
	
	if ((data = sysdiag_get_record(key)) == NULL) {
		return -1;
	}
	num = sysdiag_db_find_element(data, device);
	if (num != 0) {
		strncpy(TOPO_NAME, root_name, strlen(root_name));
		strncpy(DEVICE_NAME, device, strlen(device));
		DIAG_OPTIONS |= DIAG_CLASS | DIAG_DEVICE;
		free(data);
		data = NULL;
		return 0;
	}
	return -1;

}


/**
 *
 *
 */
int search_root(char *root_name, char *device)
{
	struct sysfs_device *dev = NULL;
	struct sysfs_root_device *root = NULL;
	
	if ((root = sysfs_open_root_device(root_name)) == NULL) {
		fprintf(stderr, "error opening root device: %s\n", root_name);
		return -1;
	}
	if (strcmp(device, root_name) == 0) {
		DIAG_OPTIONS |= DIAG_TOPO;
		strncpy(TOPO_NAME, root_name, strlen(root_name));
		sysfs_close_root_device(root);
		return 0;
	}
	if (root->devices != NULL) {
		if (search_devices(root->devices, device) == 0) {
			DIAG_OPTIONS |= DIAG_TOPO;
			strncpy(TOPO_NAME, root_name, strlen(root_name));
			sysfs_close_root_device(root);
			return 0;
		}
	}
	sysfs_close_root_device(root);
	return -1;
}

/**
 * this will search thro the database for the device information.
 *
 */
int search_class(char *class_name, char *device)
{
	char *data = NULL;
	char key[SYSFS_NAME_LEN];
	int num = 0;
	
	memset(key, 0, SYSFS_NAME_LEN);
	strncpy(key, SYSDIAG_CLASS_STR, strlen(SYSDIAG_CLASS_STR));
	strncat(key, "#", 1);
	strncat(key, SYSFS_DEVICES_NAME, strlen(SYSFS_DEVICES_NAME));
	strncat(key, "#", 1);
	strncat(key, class_name, strlen(class_name));
	num = 0;
	
	if ((data = sysdiag_get_record(key)) == NULL) {
		return -1;
	}
	num = sysdiag_db_find_element(data, device);
	if (num != 0) {
		strncpy(CLASS_NAME, class_name, strlen(class_name));
		strncpy(DEVICE_NAME, device, strlen(device));
		DIAG_OPTIONS |= DIAG_CLASS | DIAG_DEVICE;
		free(data);
		data = NULL;
		return 0;
	}
	return -1;

}
/**
 * 
 * 
 */
int old_search_class(char *class_name, char *device)
{
	struct sysfs_class *cls = NULL;
	
	if ((cls = sysfs_open_class(class_name)) == NULL) {
		fprintf(stderr, "error opening bus: %s\n", class_name);
		return -1;
	}
	if (strcmp(cls->name, device) == 0) {
		strncpy(CLASS_NAME, cls->name, strlen(cls->name));
		DIAG_OPTIONS |= DIAG_CLASS;
		sysfs_close_class(cls);
		return 0;
	}
	if (cls->devices != NULL) {
		if (search_class_devices(cls->devices, device) == 0) {
			DIAG_OPTIONS |= DIAG_CLASS | DIAG_DEVICE;
			strncpy(CLASS_NAME, cls->name, strlen(cls->name));
			sysfs_close_class(cls);
			return 0;
		}
	}
	sysfs_close_class(cls);
	return -1;
}

/**
 *
 *
 */
int search_bus(char *bus_name, char *device)
{
	char *data = NULL, *str = NULL;
	char key[SYSFS_NAME_LEN];
	int num = 0, j = 0;
	struct sysfs_bus *bus = NULL;

	if ((data = sysdiag_get_record(SYSDIAG_BUS_STR)) == NULL) {
		return -1;
	}
	num = sysdiag_db_find_element(data, device);
	if (num != 0) {
		strncpy(BUS_NAME, device, strlen(device));
		DIAG_OPTIONS |= DIAG_BUS;
		free(data);
		data = NULL;
		return 0;
	}
	free(data);
	data = NULL;
	

	memset(key, 0, SYSFS_NAME_LEN);
	strncpy(key, SYSDIAG_BUS_STR, strlen(SYSDIAG_BUS_STR));
	strncat(key, "#", 1);
	strncat(key, SYSFS_DEVICES_NAME, strlen(SYSFS_DEVICES_NAME));
	strncat(key, "#", 1);
	strncat(key, bus_name, strlen(bus_name));
	num = 0;
	
	if ((data = sysdiag_get_record(key)) == NULL) {
		return -1;
	}
	
	
	num = sysdiag_db_find_element(data, device);
	if (num != 0) {
		strncpy(BUS_NAME, bus_name, strlen(bus_name));
		strncpy(DEVICE_NAME, device, strlen(device));
		DIAG_OPTIONS |= DIAG_BUS | DIAG_DEVICE;
		free(data);
		data = NULL;
		return 0;
	}
	free(data);
	data = NULL;
	
	memset(key, 0, SYSFS_NAME_LEN);
	strncpy(key, SYSDIAG_BUS_STR, strlen(SYSDIAG_BUS_STR));
	strncat(key, "#", 1);
	strncat(key, SYSFS_DRIVERS_NAME, strlen(SYSFS_DRIVERS_NAME));
	strncat(key, "#", 1);
	strncat(key, bus_name, strlen(bus_name));
	num = 0;
	
	if ((data = sysdiag_get_record(key)) == NULL) {
		return -1;
	}
	num = sysdiag_db_find_element(data, device);
	if (num != 0) {
		strncpy(BUS_NAME, bus_name, strlen(bus_name));
		strncpy(DRIVER_NAME, device, strlen(device));
		DIAG_OPTIONS |= DIAG_BUS | DIAG_DRIVER;
		free(data);
		data = NULL;
		return 0;
	}
	free(data);
	data = NULL;
	return -1;
}

int find_dev(char *device)
{
	char *data = NULL, *str = NULL;
	int j = 0,  num = 0;
	
	data = sysdiag_get_record(SYSDIAG_BUS_STR);
	if (data != NULL) {
		if ((num = sysdiag_db_num_of_elements(data)) != 0) {
			for (j = 1; j <= num; j++) {
				if ((str = sysdiag_db_get_str(data, j)) != NULL) {
					if (search_bus(str, device) == 0) {
						free(data);
						return 0;
					}
				}
			}
		}
		free(data);
		data = NULL;
	}
	
	data = sysdiag_get_record(SYSDIAG_CLASS_STR);
	if (data != NULL) {
		num = sysdiag_db_find_element(data, device);
		if (num != 0) {
			strncpy(CLASS_NAME, device, strlen(device));
			DIAG_OPTIONS |= DIAG_CLASS;
			free(data);
			data = NULL;
			return 0;
		}
		if ((num = sysdiag_db_num_of_elements(data)) != 0) {
			for (j = 1; j <= num; j++) {
				if ((str = sysdiag_db_get_str(data, j)) != NULL) {
					if (search_class(str, device) == 0) {
						free(data);
						return 0;
					}
				}
			}
		}
		free(data);
		data = NULL;
	}

	data = sysdiag_get_record(SYSFS_DEVICES_NAME);
	if (data != NULL) {
		if ((num = sysdiag_db_num_of_elements(data)) != 0) {
			for (j = 1; j <= num; j++) {
				if ((str = sysdiag_db_get_str(data, j)) != NULL) {
					if (search_root(str, device) == 0) {
						free(data);
						return 0;
					}
				}
			}
		}
		free(data);
		data = NULL;
	}
	return -1;
	
}

int find_device(int argc, char *argv[])
{
	int i;
	if (argc == 0) return 0;

	for (i = 0; i < argc ; i++) {
		if (find_dev(argv[i]) != 0) {
			fprintf(stderr, "device not found: %s\n", argv[i]);
			return -1;
		}
	}
	return 0;	
}


/**
 * main function.
 *
 * sysdiag <options> [bus|class] [device]
 *
 * sysdiag -a eth0
 * sysdiag -i eth0
 * sysdiag net eth0
 *
 */
int main(int argc, char *argv[])
{
	int opt;
	extern char *optarg;
	extern int optind, opterr, optopt;
	DIAG_LEVEL = 0;
	DIAG_OPTIONS = 0;
	diag_list = NULL;

	memset(BUS_NAME, 0, SYSFS_NAME_LEN);
	memset(TOPO_NAME, 0, SYSFS_NAME_LEN);
	memset(CLASS_NAME, 0, SYSFS_NAME_LEN);
	memset(BLOCK_NAME, 0, SYSFS_NAME_LEN);
	memset(DEVICE_NAME, 0, SYSFS_NAME_LEN);
	memset(DRIVER_NAME, 0, SYSFS_NAME_LEN);

	while ((opt = getopt(argc,argv,CMD_OPTIONS)) != -1) {
		switch (opt) {
			case 'b':
				if (optarg == 0) {
					usage();
					exit(0);
				}
				strncpy(BUS_NAME, optarg, strlen(optarg));
				DIAG_OPTIONS |= DIAG_BUS;
				break;
			case 'c':
				if (optarg == 0) {
					usage();
					exit(0);
				}
				strncpy(CLASS_NAME, optarg, strlen(optarg));
				DIAG_OPTIONS |= DIAG_CLASS;
				break;
			case 'i':
				DIAG_OPTIONS |= DIAG_INQ;
				break;
			case 'd':
				DIAG_OPTIONS |= DIAG_DIAG_INFO;
				
				break;

			case 'L':
				DIAG_OPTIONS |= DIAG_SHOW_OLD_LOGS;
			case 'l':
				DIAG_OPTIONS |= DIAG_LOGS;
				break;
			case 'p':
				DIAG_OPTIONS |= DIAG_PAUSE;
				break;
			case 's':
				DIAG_OPTIONS |= DIAG_STAT;
				break;
			case 't':
				DIAG_OPTIONS |= DIAG_TEST;
				break;
			case 'r':
				if (optarg == 0) {
					usage();
					exit(0);
				}
				strncpy(TOPO_NAME, optarg, strlen(optarg));
				DIAG_OPTIONS |= DIAG_TOPO;
				break;
			case 'R':
				DIAG_OPTIONS |= DIAG_DB_RELOAD;
				break;
			case 'v':
				DIAG_OPTIONS |= DIAG_SHOW_VALUE 
					     | DIAG_LIST;
				DIAG_LEVEL++;
				break;
			case 'x':
				DIAG_OPTIONS |= DIAG_ALL;
				break;
			case 'h':
				usage();
				exit(0);
			default:
				usage();
				exit(0);
		}
	}
	argc -= optind;
	argv += optind;
	
	if (parse_main() == -1) {
		return -1;
		
	}
	if (DIAG_OPTIONS & DIAG_DIAG_INFO) {
		show_diag_info(argv[0]);
	}
	if (setup_sysfs_paths() != 0) {
		fprintf(stderr, "sysfs not found\n");
		sysdiag_db_close();
		return -1;
	}
	
	if (find_device(argc,argv) == -1) {
		sysdiag_db_close();
		return -1;
	}
	if ((DIAG_OPTIONS & DIAG_INQ)  ||
	    (DIAG_OPTIONS & DIAG_TEST) ||
	    (DIAG_OPTIONS & DIAG_LOGS) ||
	    (DIAG_OPTIONS & DIAG_STAT)) {
		run_diag();
//		if (run_diag() != 0)
//			fprintf(stdout, "failed to run diagnostics on the "
//					"requested device\n");
	} else { 
		if (DIAG_OPTIONS & DIAG_ALL) {
			usage();
			return 0;
		}
		if (!(DIAG_OPTIONS & DIAG_DIAG_INFO) && 
				!(DIAG_OPTIONS & DIAG_DB_RELOAD)) {
			DIAG_OPTIONS |= DIAG_LIST;
			if ((DEVICE_NAME[0] == 0 ) && (DRIVER_NAME[0] == 0)) {
				DIAG_OPTIONS |= DIAG_DEVICE | DIAG_DRIVER;
			}
		}
	}
	if (DIAG_OPTIONS & DIAG_LIST) {
		list_main();
	}

	sysdiag_db_close();
	return 0;

} /* end main */
